var debugLogger = null

function initLogger(logger) {
  debugLogger = logger
}

function debugLog(msg, name, lvl) {
  if (debugLogger) {
    debugLogger(msg, name, lvl)
  }
}

const commonUselessAnswerParts = [
  'A helyes válasz az ',
  'A helyes válasz a ',
  'A helyes válaszok: ',
  'A helyes válaszok:',
  'A helyes válasz: ',
  'A helyes válasz:',
  'The correct answer is:',
  "'",
]
const commonUselessStringParts = [',', '\\.', ':', '!', '\\+', '\\s*\\.']
const specialChars = ['&', '\\+']
const lengthDiffMultiplier = 10 /* Percent minus for length difference */
const minMatchAmmount = 60 /* Minimum ammount to consider that two questions match during answering */

const assert = (val) => {
  if (!val) {
    throw new Error('Assertion failed')
  }
}

// ---------------------------------------------------------------------------------------------------------
// String Utils
// ---------------------------------------------------------------------------------------------------------

// Exported
// ---------------------------------------------------------------------------------------------------------
function getSubjNameWithoutYear(subjName) {
  let t = subjName.split(' - ')
  if (t[0].match(/^[0-9]{4}\/[0-9]{2}\/[0-9]{1}$/i)) {
    return t[1] || subjName
  } else {
    return subjName
  }
}

// Not exported
// ---------------------------------------------------------------------------------------------------------
function removeStuff(value, removableStrings, toReplace) {
  removableStrings.forEach((removableString) => {
    var regex = new RegExp(removableString, 'g')
    value = value.replace(regex, toReplace || '')
  })
  return value
}

// removes whitespace from begining and and, and replaces multiple spaces with one space
function removeUnnecesarySpaces(toremove) {
  assert(toremove)

  toremove = normalizeSpaces(toremove)
  while (toremove.includes('  ')) {
    toremove = toremove.replace(/ {2}/g, ' ')
  }
  return toremove.trim()
}

// simplifies a string for easier comparison
function simplifyStringForComparison(value) {
  assert(value)

  value = removeUnnecesarySpaces(value).toLowerCase()
  return removeStuff(value, commonUselessStringParts)
}

function removeSpecialChars(value) {
  assert(value)

  return removeStuff(value, specialChars, ' ')
}

// damn nonbreaking space
function normalizeSpaces(input) {
  assert(input)

  return input.replace(/\s/g, ' ')
}

function compareString(s1, s2) {
  if (!s1 || !s2) {
    if (!s1 && !s2) {
      return 100
    } else {
      return 0
    }
  }

  s1 = simplifyStringForComparison(s1).split(' ')
  s2 = simplifyStringForComparison(s2).split(' ')
  var match = 0
  for (var i = 0; i < s1.length; i++) {
    if (s2.includes(s1[i])) {
      match++
    }
  }
  var percent = Math.round(((match / s1.length) * 100).toFixed(2)) // matched words percent
  var lengthDifference = Math.abs(s2.length - s1.length)
  percent -= lengthDifference * lengthDiffMultiplier
  if (percent < 0) {
    percent = 0
  }
  return percent
}

function answerPreProcessor(value) {
  assert(value)

  return removeStuff(value, commonUselessAnswerParts)
}

// 'a. pécsi sör' -> 'pécsi sör'
function removeAnswerLetters(value) {
  assert(value)

  let val = value.split('. ')
  if (val[0].length < 2 && val.length > 1) {
    val.shift()
    return val.join(' ')
  } else {
    return value
  }
}

function simplifyQA(value, mods) {
  if (!value) {
    return
  }

  const reducer = (res, fn) => {
    return fn(res)
  }

  return mods.reduce(reducer, value)
}

function simplifyAnswer(value) {
  return simplifyQA(value, [
    removeSpecialChars.bind(this),
    removeUnnecesarySpaces.bind(this),
    answerPreProcessor.bind(this),
    removeAnswerLetters.bind(this),
  ])
}

function simplifyQuestion(question) {
  if (typeof q === 'string') {
    return simplifyQA(question, [
      removeSpecialChars.bind(this),
      removeUnnecesarySpaces.bind(this),
    ])
  } else {
    question.Q = simplifyQA(question.Q, [
      removeSpecialChars.bind(this),
      removeUnnecesarySpaces.bind(this),
    ])
    question.A = simplifyQA(question.A, [
      removeSpecialChars.bind(this),
      removeUnnecesarySpaces.bind(this),
    ])
    return question
  }
}

// ---------------------------------------------------------------------------------------------------------
// Question
// ---------------------------------------------------------------------------------------------------------

function createQuestion(question, answer, data) {
  return {
    Q: simplifyQuestion(question),
    A: simplifyAnswer(answer),
    data,
  }
}

function compareImage(data, data2) {
  return compareString(data.images.join(' '), data2.images.join(' '))
}

function compareData(q1, q2) {
  try {
    if (q1.data.type === q2.data.type) {
      let dataType = q1.data.type
      if (dataType === 'simple') {
        return -1
      } else if (dataType === 'image') {
        return compareImage(q1.data, q2.data)
      } else {
        debugLog(`Unhandled data type ${dataType}`, 'Compare question data', 1)
        debugLog(q1, 'Compare question data', 2)
      }
    } else {
      return 0
    }
  } catch (error) {
    debugLog('Error comparing data', 'Compare question data', 1)
    debugLog(error.message, 'Compare question data', 1)
    debugLog(error, 'Compare question data', 2)
  }
  return 0
}

function compareQuestion(q1, q2) {
  return compareString(q1.Q, q2.Q)
}

function compareAnswer(q1, q2) {
  return compareString(q1.A, q2.A)
}

function compareQuestionObj(q1, q1subjName, q2, q2subjName, data) {
  assert(data)
  assert(q1)
  assert(typeof q1 === 'object')
  assert(q2)
  let qObj

  if (typeof q2 === 'string') {
    qObj = {
      Q: q2,
      data: data,
    }
  } else {
    qObj = q2
  }

  const qMatch = compareQuestion(q1, qObj)
  const aMatch = compareAnswer(q1, qObj)
  // -1 if botth questions are simple
  const dMatch = compareData(q1, qObj)

  let avg = -1
  if (qObj.A) {
    if (dMatch === -1) {
      avg = (qMatch + aMatch) / 2
    } else {
      avg = (qMatch + aMatch + dMatch) / 3
    }
  } else {
    if (dMatch === -1) {
      avg = qMatch
    } else {
      avg = (qMatch + dMatch) / 2
    }
  }

  return {
    qMatch: qMatch,
    aMatch: aMatch,
    dMatch: dMatch,
    matchedSubjName: q2subjName,
    avg: avg,
  }
}

function questionToString(question) {
  const { Q, A, data } = question

  if (data.type !== 'simple') {
    return '?' + Q + '\n!' + A + '\n>' + JSON.stringify(data)
  } else {
    return '?' + Q + '\n!' + A
  }
}

// ---------------------------------------------------------------------------------------------------------
// Subject
// ---------------------------------------------------------------------------------------------------------
function searchQuestion(subj, question, questionData, subjName) {
  assert(question)

  var result = []
  subj.Questions.forEach((currentQuestion) => {
    let percent = compareQuestionObj(
      currentQuestion,
      subjName,
      question,
      subj.Name,
      questionData
    )
    if (percent.avg > minMatchAmmount) {
      result.push({
        question: currentQuestion,
        match: percent.avg,
        detailedMatch: percent,
      })
    }
  })

  result = result.sort((q1, q2) => {
    return q1.match < q2.match
  })

  return result
}

function subjectToString(subj) {
  const { Questions, Name } = subj

  var result = []
  Questions.forEach((question) => {
    result.push(questionToString(question))
  })

  return '+' + Name + '\n' + result.join('\n')
}

// ---------------------------------------------------------------------------------------------------------
// QuestionDB
// ---------------------------------------------------------------------------------------------------------
function addQuestion(data, subj, question) {
  debugLog('Adding new question with subjName: ' + subj, 'qdb add', 1)
  debugLog(question, 'qdb add', 3)
  assert(data)
  assert(subj)
  assert(question)
  assert(typeof question === 'object')
  let result = []

  var i = 0
  while (
    i < data.length &&
    !subj
      .toLowerCase()
      .includes(getSubjNameWithoutYear(data[i].Name).toLowerCase())
  ) {
    i++
  }

  if (i < data.length) {
    debugLog('Adding new question to existing subject', 'qdb add', 1)
    result = [...data]
    result[i].Questions = [...data[i].Questions, question]
  } else {
    debugLog('Creating new subject for question', 'qdb add', 1)
    result = [
      ...data,
      {
        name: subj,
        Questions: [question],
      },
    ]
  }

  return result
}

function searchData(data, question, subjName, questionData) {
  assert(data)
  assert(question)
  debugLog('Searching for question', 'qdb search', 1)
  debugLog('Question:', 'qdb search', 2)
  debugLog(question, 'qdb search', 2)
  debugLog(`Subject name: ${subjName}`, 'qdb search', 2)
  debugLog('Data:', 'qdb search', 2)
  debugLog(questionData || question.data, 'qdb search', 2)

  if (!questionData) {
    questionData = question.data || { type: 'simple' }
  }
  if (!subjName) {
    subjName = ''
    debugLog('No subject name as param!', 'qdb search', 1)
  }
  question = simplifyQuestion(question)

  let result = []
  data.forEach((subj) => {
    if (
      subjName
        .toLowerCase()
        .includes(getSubjNameWithoutYear(subj.Name).toLowerCase())
    ) {
      debugLog(`Searching in ${subj.Name} `, 2)
      result = result.concat(
        searchQuestion(subj, question, questionData, subjName)
      )
    }
  })

  // FIXME: try to remove this? but this is also a good backup plan so idk
  if (result.length === 0) {
    debugLog(
      'Reqults length is zero when comparing names, trying all subjects',
      'qdb search',
      1
    )
    data.forEach((subj) => {
      result = result.concat(
        searchQuestion(subj, question, questionData, subjName)
      )
    })
    if (result.length > 0) {
      debugLog(
        `FIXME: '${subjName}' gave no result but '' did!`,
        'qdb search',
        1
      )
      console.error(`FIXME: '${subjName}' gave no result but '' did!`)
    }
  }

  result = result.sort((q1, q2) => {
    return q1.match < q2.match
  })

  debugLog(`QDB search result length: ${result.length}`, 'qdb search', 1)
  return result
}

function addSubject(data, subj) {
  assert(data)
  assert(subj)

  var i = 0
  while (i < length && subj.Name !== data[i].Name) {
    i++
  }

  if (i < length) {
    return data.map((currSubj, j) => {
      if (j === i) {
        return {
          ...currSubj,
          Questions: [...currSubj.Questions, ...subj.Questions],
        }
      } else {
        return currSubj
      }
    })
  } else {
    return [...data, subj]
  }
}

function dataToString(data) {
  var result = []
  data.forEach((subj) => {
    result.push(subjectToString(subj))
  })
  return result.join('\n\n')
}

module.exports = {
  minMatchAmmount,
  initLogger,
  getSubjNameWithoutYear,
  createQuestion,
  addQuestion,
  addSubject,
  searchData,
  dataToString,
}
